import os
from tempfile import mkdtemp

import biocframe
import dolomite_sce
import numpy
from dolomite_base import read_object, save_object
from singlecellexperiment import SingleCellExperiment
from summarizedexperiment import SummarizedExperiment


def test_stage_sce_basic():
    x = numpy.random.rand(1000, 200)
    se = SingleCellExperiment({"counts": x})

    dir = os.path.join(mkdtemp(), "sce_simple")
    save_object(se, dir)

    roundtrip = read_object(dir)
    assert isinstance(roundtrip, SingleCellExperiment)
    ass = roundtrip.assay("counts")
    assert ass.shape == (1000, 200)

    # Works with multiple assays.
    x2 = (numpy.random.rand(1000, 200) * 10).astype(numpy.int32)
    se = SingleCellExperiment(
        {"logcounts": x, "counts": x2}, main_experiment_name="aaron's secret modality"
    )

    dir = os.path.join(mkdtemp(), "sce_simple2")
    save_object(se, dir)

    roundtrip = read_object(dir)
    assert roundtrip.assay_names == ["logcounts", "counts"]
    assert se.main_experiment_name == roundtrip.main_experiment_name


def test_stage_sce_with_dimdata_with_names():
    x = numpy.random.rand(1000, 200)
    se = SingleCellExperiment(
        assays={"counts": x},
        row_data=biocframe.BiocFrame(row_names=["gene" + str(i) for i in range(1000)]),
        column_data=biocframe.BiocFrame(
            row_names=["cell" + str(i) for i in range(200)]
        ),
    )

    dir = os.path.join(mkdtemp(), "sce_dimdata2")
    save_object(se, dir)

    roundtrip = read_object(dir)
    assert isinstance(roundtrip, SingleCellExperiment)
    assert se.row_data.row_names == roundtrip.row_data.row_names
    assert se.column_data.row_names == roundtrip.column_data.row_names


def test_stage_sce_with_rdims():
    x = numpy.random.rand(1000, 200)
    se = SingleCellExperiment(
        assays={"counts": x},
        row_data=biocframe.BiocFrame(
            {"foo": numpy.random.rand(1000), "bar": numpy.random.rand(1000)}
        ),
        column_data=biocframe.BiocFrame(
            {"whee": numpy.random.rand(200), "stuff": numpy.random.rand(200)}
        ),
        reduced_dims={"tsnooch": numpy.random.rand(200, 4)},
    )

    dir = os.path.join(mkdtemp(), "sce_dimdata")
    save_object(se, dir)

    roundtrip = read_object(dir)
    assert isinstance(roundtrip, SingleCellExperiment)
    assert numpy.allclose(se.row_data["foo"], roundtrip.row_data["foo"])
    assert numpy.allclose(se.column_data["stuff"], roundtrip.column_data["stuff"])
    assert se.get_reduced_dim_names() == roundtrip.get_reduced_dim_names()
    assert numpy.allclose(
        se.reduced_dim("tsnooch"), numpy.array(roundtrip.reduced_dim("tsnooch"))
    )


def test_stage_sce_with_rdims_and_alts():
    x = numpy.random.rand(1000, 200)
    se = SingleCellExperiment(
        assays={"counts": x},
        row_data=biocframe.BiocFrame(
            {"foo": numpy.random.rand(1000), "bar": numpy.random.rand(1000)}
        ),
        column_data=biocframe.BiocFrame(
            {"whee": numpy.random.rand(200), "stuff": numpy.random.rand(200)}
        ),
        reduced_dims={"tsnooch": numpy.random.rand(200, 4)},
        alternative_experiments={
            "useless_modality": SummarizedExperiment(
                {"counts": numpy.random.rand(100, 200)}
            )
        },
    )

    dir = os.path.join(mkdtemp(), "sce_dimdata")
    save_object(se, dir)

    roundtrip = read_object(dir)
    assert isinstance(roundtrip, SingleCellExperiment)
    assert numpy.allclose(se.row_data["foo"], roundtrip.row_data["foo"])
    assert numpy.allclose(se.column_data["stuff"], roundtrip.column_data["stuff"])
    assert se.get_reduced_dim_names() == roundtrip.get_reduced_dim_names()
    assert numpy.allclose(
        se.reduced_dim("tsnooch"), numpy.array(roundtrip.reduced_dim("tsnooch"))
    )
    assert (
        se.get_alternative_experiment_names()
        == roundtrip.get_alternative_experiment_names()
    )
    assert numpy.allclose(
        se.alternative_experiment("useless_modality").assay("counts"),
        numpy.array(
            roundtrip.alternative_experiment("useless_modality").assay("counts")
        ),
    )


def test_stage_sce_with_other_meta():
    x = numpy.random.rand(1000, 200)
    se = SingleCellExperiment(assays={"counts": x}, metadata={"YAY": 2, "FOO": "a"})

    dir = os.path.join(mkdtemp(), "sce_other_meta")
    save_object(se, dir)

    roundtrip = read_object(dir)
    assert roundtrip.metadata == se.metadata


def test_empty_sce():
    se = SingleCellExperiment(assays={}, metadata={"YAY": 2, "FOO": "a"})

    dir = os.path.join(mkdtemp(), "sce_other_meta2")
    save_object(se, dir)

    roundtrip = read_object(dir)
    assert roundtrip.metadata == se.metadata
    assert len(se.get_assay_names()) == len(roundtrip.get_assay_names())


def test_empty_dimnames():
    se = SingleCellExperiment(
        assays={},
        row_data=biocframe.BiocFrame(row_names=["gene" + str(i) for i in range(1000)]),
        column_data=biocframe.BiocFrame(
            row_names=["cell" + str(i) for i in range(200)]
        ),
    )

    dir = os.path.join(mkdtemp(), "sce_dimdata3")
    save_object(se, dir)

    roundtrip = read_object(dir)
    assert isinstance(roundtrip, SingleCellExperiment)
    assert se.row_data.row_names == roundtrip.row_data.row_names
    assert se.column_data.row_names == roundtrip.column_data.row_names
    assert len(se.get_assay_names()) == len(roundtrip.get_assay_names())



import json
import os

import dolomite_base as dl
import dolomite_se as dlse
from dolomite_base.read_object import read_object_registry
from singlecellexperiment import SingleCellExperiment

read_object_registry["single_cell_experiment"] = (
    "dolomite_sce.read_single_cell_experiment"
)


def read_single_cell_experiment(
    path: str, metadata: dict, **kwargs
) -> SingleCellExperiment:
    """Load a
    :py:class:`~singlecellexperiment.SingleCellExperiment.SingleCellExperiment`
    from its on-disk representation.

    This method should generally not be called directly but instead be invoked by
    :py:meth:`~dolomite_base.read_object.read_object`.

    Args:
        path:
            Path to the directory containing the object.

        metadata:
            Metadata for the object.

        kwargs:
            Further arguments.

    Returns:
        A
        :py:class:`~singlecellexperiment.SingleCellExperiment.SingleCellExperiment`
        with file-backed arrays in the assays.
    """

    # see comments in read_ranged_summarized_experiment in dolomite_se.
    rse = dlse.read_ranged_summarized_experiment(path, metadata=metadata, **kwargs)

    _main_expt_name = None
    if "main_experiment_name" in metadata["single_cell_experiment"]:
        _main_expt_name = metadata["single_cell_experiment"]["main_experiment_name"]

    sce = SingleCellExperiment(
        assays=rse.get_assays(),
        row_data=rse.get_row_data(),
        column_data=rse.get_column_data(),
        row_ranges=rse.get_row_ranges(),
        metadata=rse.get_metadata(),
        main_experiment_name=_main_expt_name,
    )

    _rdim_path = os.path.join(path, "reduced_dimensions")
    if os.path.exists(_rdim_path):
        _rdims = {}

        with open(os.path.join(_rdim_path, "names.json"), "r") as handle:
            _rdim_names = json.load(handle)

        for _aidx, _aname in enumerate(_rdim_names):
            _rdim_read_path = os.path.join(_rdim_path, str(_aidx))

            try:
                _rdims[_aname] = dl.alt_read_object(_rdim_read_path, **kwargs)
            except Exception as ex:
                raise RuntimeError(
                    f"failed to load reduced dimension '{_aname}' from '{path}'; "
                    + str(ex)
                )

        sce = sce.set_reduced_dims(_rdims)

    _alt_path = os.path.join(path, "alternative_experiments")
    if os.path.exists(_alt_path):
        _alts = {}

        with open(os.path.join(_alt_path, "names.json"), "r") as handle:
            _alt_names = json.load(handle)

        for _aidx, _aname in enumerate(_alt_names):
            _alt_read_path = os.path.join(_alt_path, str(_aidx))

            try:
                _alts[_aname] = dl.alt_read_object(_alt_read_path, **kwargs)
            except Exception as ex:
                raise RuntimeError(
                    f"failed to load alternative experiment '{_aname}' from '{path}'; "
                    + str(ex)
                )

        sce = sce.set_alternative_experiments(_alts)

    return sce



import json
import os

import dolomite_base as dl
import dolomite_se as dlse
from singlecellexperiment import SingleCellExperiment
from summarizedexperiment import RangedSummarizedExperiment


@dl.save_object.register
@dl.validate_saves
def save_single_cell_experiment(
    x: SingleCellExperiment,
    path: str,
    data_frame_args: dict = None,
    assay_args: dict = None,
    rdim_args: dict = None,
    alt_expts_args: dict = None,
    **kwargs,
):
    """Method for saving
    :py:class:`~singlecellexperiment.SingleCellExperiment.SingleCellExperiment`
    objects to their corresponding file representations, see
    :py:meth:`~dolomite_base.save_object.save_object` for details.

    Args:
        x:
            Object to be staged.

        path:
            Path to a directory in which to save ``x``.

        data_frame_args:
            Further arguments to pass to the ``save_object`` method for the
            row/column data.

        assay_args:
            Further arguments to pass to the ``save_object`` method for the
            assays.

        rdim_args:
            Further arguments to pass to the ``save_object`` method for the
            reduced dimensions.

        alt_expts_args:
            Further arguments to pass to the ``save_object`` method for the
            alternative experiments.

        kwargs:
            Further arguments.

    Returns:
        ``x`` is saved to path.
    """
    if data_frame_args is None:
        data_frame_args = {}

    if assay_args is None:
        assay_args = {}

    if rdim_args is None:
        rdim_args = {}

    if alt_expts_args is None:
        alt_expts_args = {}

    # see comments in save_ranged_summarized_experiment in dolomite_se.
    dlse.save_ranged_summarized_experiment(
        x, path, data_frame_args=data_frame_args, assay_args=assay_args, **kwargs
    )

    # Modify OBJECT
    _info = dl.read_object_file(path)
    _info["single_cell_experiment"] = {"version": "1.0"}
    if x.get_main_experiment_name() is not None:
        _info["single_cell_experiment"]["main_experiment_name"] = str(
            x.get_main_experiment_name()
        )
    dl.save_object_file(path, "single_cell_experiment", _info)

    # save rdims
    _rdim_names = x.get_reduced_dim_names()
    if len(_rdim_names) > 0:
        _rdim_path = os.path.join(path, "reduced_dimensions")
        os.mkdir(_rdim_path)

        with open(os.path.join(_rdim_path, "names.json"), "w") as handle:
            json.dump(_rdim_names, handle)

        for _aidx, _aname in enumerate(_rdim_names):
            _rdim_save_path = os.path.join(_rdim_path, str(_aidx))
            try:
                dl.alt_save_object(
                    x.reduced_dim(_aname), path=_rdim_save_path, **rdim_args, **kwargs
                )
            except Exception as ex:
                raise RuntimeError(
                    "failed to stage reduced dimension '"
                    + _aname
                    + "' for "
                    + str(type(x))
                    + "; "
                    + str(ex)
                )

    # save alt expts.
    _alt_names = x.get_alternative_experiment_names()
    if len(_alt_names) > 0:
        _alt_path = os.path.join(path, "alternative_experiments")
        os.mkdir(_alt_path)

        with open(os.path.join(_alt_path, "names.json"), "w") as handle:
            json.dump(_alt_names, handle)

        for _aidx, _aname in enumerate(_alt_names):
            _alt_save_path = os.path.join(_alt_path, str(_aidx))
            try:
                dl.alt_save_object(
                    x.alternative_experiment(_aname),
                    path=_alt_save_path,
                    **alt_expts_args,
                    **kwargs,
                )
            except Exception as ex:
                raise RuntimeError(
                    "failed to stage alternative experiment '"
                    + _aname
                    + "' for "
                    + str(type(x))
                    + "; "
                    + str(ex)
                )
    return
